/*
 * invest.c
 *
 * This program takes an initial investment amount, an expected annual
 * return rate, and the number of years, and calculates the growth of
 * the investment.  The main point of this program, though, is to
 * demonstrate some C syntax.  For more on C, see 
 *
 * Essential C:  http://cslibrary.stanford.edu/101/
 * Also:  http://publications.gbdirect.co.uk/c_book/
 * 
 * HISTORY:
 *
 * 12/31/2010  Created by Nelson Rosa
 * 1/3/2011    Modified by Kevin Lynch
 *
 */

// Comments:  everything starting /* and ending */ is a comment.
// Also, things after // on the same line are treated as comments by
// most C compilers, although this is not included in the ANSI C standard.

/******************************************************************************
 * PREPROCESSOR COMMANDS
 *
 * Library "include" files, constants, etc. Essential C, pp. 29-31, 41-45.
 ******************************************************************************/
 
#include <stdlib.h>  	// memory allocation
#include <stdio.h>      // input/output, like scanf and printf
#include <math.h>    	// common, but we don't need the math library here

#include "pic32_additions.h" // our own header file

#define MAX_YEARS 100   // Calculate no more than 100 years.

/******************************************************************************
 * STRUCT, AND DEFINING A STRUCT VARIABLE TYPE
 *
 * It is convenient to package a number of variables that are always used
 * together into a struct.  Essential C, p. 15, 23.
 ******************************************************************************/

struct InvestmentStruct {
  double inv0;    	// initial investment
  double growth;	// growth rate, where 1.0 indicates zero growth
  int years;		// number of years to track
};

typedef struct InvestmentStruct InvParams;	// the new variable type

/******************************************************************************
 * GLOBAL VARIABLES
 *
 * Global variables are defined outside a function.  They are useful
 * when a variable, such as a flag, must be accessed by a number of
 * functions.  They defeat modularity, though, so they should be used
 * sparingly.  Variable types are discussed in Essential C, pp. 3-5.
 ******************************************************************************/

// no global variables in this program

/******************************************************************************
 * FUNCTION PROTOTYPES
 *
 * Function prototypes indicate what variable types a function takes
 * as inputs, and what type it returns (Essential C, pp. 24-26, 29).  The
 * compiler needs this info before it first encounters a call to that
 * function.  These prototypes are sometimes placed in a .h "header"
 * file.  When your program extends over multiple files, putting a
 * "static" before the function type in front ensures it is only
 * available to functions in this file.
 ******************************************************************************/

int getUserInput(InvParams *invp);  // pointers, * and &.  Essential C
				    // pp. 17-20, 26-28, 33-40
double *allocateInvestmentArray(int years);
void calculateGrowth(InvParams *invp, double *array);
void sendOutput(InvParams *invp, double *array);
void freeInvestmentArray(double *array);

/******************************************************************************
 * MAIN FUNCTION
 *
 * Each program needs a single main function (Essential C, p. 29).
 * This one takes no inputs; argc and argv are common for command line.
 ******************************************************************************/

int main(void) {

  InvParams inv;	// these variables are local to main.
  double *invarray;	// invarray is a pointer to type double.  We
			// have not allocated the memory for the
			// actual array yet.  Alternatively, we could
			// statically allocate with "double invarray[100]".
			// Then we'd have 100 doubles set aside, with
			// invarray being a pointer to the first
			// element, i.e., &invarray[0].  Essential C,
			// pp. 15-16, 33-40.
  int num=5, den=2;	// num, den, ratio are for a simple casting example
  float ratio;
  
  // initialize the UART1 serial port
  initUart1();
  
  while(getUserInput(&inv)) {	// while loop, Essential C p. 12
    invarray = allocateInvestmentArray(inv.years);  // . references struct field
    calculateGrowth(&inv,invarray);	// & gives the address of inv
    sendOutput(&inv,invarray);
    freeInvestmentArray(invarray);
  }

  // gratuitous casting example, Essential C pp. 6-7
  ratio = num/den;
  printf("\nCasting example.\nRatio, %1d/%1d:  %5.2f\n",num,den,ratio);
  ratio = num/((float) den);
  printf("Ratio, %1d/((float) %1d):  %5.2f\n",num,den,ratio);
  ratio = ((float) num)/((float) den);
  printf("Ratio, ((float) %1d)/((float) %1d):  %5.2f\n",num,den,ratio);
}

/* getUserInput
 * 
 * This reads the user's input into the struct pointed at by invp, 
 * and returns TRUE if the input is valid, FALSE if not.
 */

int getUserInput(InvParams *invp) {

  int valid = 0;  	// Used as a boolean.  Zero is logical FALSE; any nonzero
		// is TRUE.  Essential C, p. 7.
  char input[256]; // Destination array

  // gets, printf, and scanf are in stdio.h, Essential C pp. 42-43.
  printf("Enter the initial investment, growth rate, number of years: ");
  gets(input);
  sscanf(input, "%lf %lf %d", &(invp->inv0), &(invp->growth), &(invp->years));
  printf("\r\n");

  // Relational and logical operators, Essential C pp. 8-9.
  // Use -> to reference struct field, not ., when you have a struct ptr.
  valid = (invp->inv0 > 0) && (invp->growth > 0) &&
    (invp->years > 0) && (invp->years <= MAX_YEARS);

  printf("Valid inputs?  %d\n",valid);

  // if-else, Essential C p. 11.
  if (!valid) { // ! is logical NOT, Essential C p. 9.
    printf("Invalid input; exiting.\n");
  }
  return(valid);
}

/* allocateInvestmentArray
 *
 * This dynamically allocates the memory for the investment array on
 * the "heap."  Contrast with static allocation.  Code below uses
 * calloc; see also malloc.  Essential C pp. 38-40, 45.
 */

double *allocateInvestmentArray(int years) {
  return((double *) calloc(years+1, sizeof(double)));
}

/* calculateGrowth
 *
 * This fills array with the investment over the years, given the
 * parameters in *invp.
 */

void calculateGrowth(InvParams *invp, double *array) {

  int i;

  array[0] = invp->inv0;
  for (i=1; i <= invp->years; i++) {  // for loop, Essential C p. 13
                                      // i++ is i=i+1, Essential C pp. 7-8
    array[i] = invp->growth*array[i-1];
  }
}

/* sendOutput
 *
 * This function prints the results.
 */

void sendOutput(InvParams *invp, double *array) {

  int i;

  printf("\nRESULTS:\n\n");
  for (i=0; i<=invp->years; i += 1) {  // same as i=i+1, Essential C p. 10
    printf("Year %3d:  %10.2f\n",i,array[i]);	// formatted output
  }
  printf("\n");
}

/* freeInvestmentArray
 *
 * This returns the memory to the heap, making it available for other
 * use.  Essential C, pp. 38-40, 45.
 */

void freeInvestmentArray(double *array) {
  free(array);
}
